import linien

from time import sleep
from socket import gaierror
from plumbum import colors
from traceback import print_exc

from linien.config import SERVER_PORT
from linien.client.utils import run_server
from linien.client.exceptions import GeneralConnectionErrorException, \
    InvalidServerVersionException
from linien.communication.client import BaseClient


class Connection(BaseClient):
    def __init__(self, host, user=None, password=None):
        self.user = user
        self.password = password

        if host in ('localhost', '127.0.0.1'):
            # RP is configured such that "localhost" doesn't point to
            # 127.0.0.1 in all cases
            host = '127.0.0.1'
        else:
            assert user and password

        super().__init__(host, SERVER_PORT, True)

        self.control = self.connection.root

    def connect(self, host, port, use_parameter_cache):
        self.connection = None

        i = -1

        while True:
            i += 1

            try:
                print('try to connect to %s:%s' % (host, port))
                self._connect(host, port, use_parameter_cache)
                break
            except gaierror:
                # host not found
                print(colors.red | 'Error: host %s not found' % host)
                break
            except Exception as e:
                if i == 0:
                    print('server is not running. Launching it!')
                    run_server(host, self.user, self.password)
                    sleep(3)
                else:
                    if i < 20:
                        print('server still not running, waiting (this may take some time)...')
                        sleep(1)
                    else:
                        print_exc()
                        print(colors.red | 'Error: connection to the server could not be established')
                        break

        if self.connection is None:
            raise GeneralConnectionErrorException()

        # now check that the remote version is the same as ours
        remote_version = self.connection.root.exposed_get_server_version()
        client_version = linien.__version__

        if remote_version != client_version:
            raise InvalidServerVersionException(client_version, remote_version)

        print(colors.green | 'connected established!')

    def disconnect(self):
        self.connection.close()


from PyQt5 import QtGui, QtWidgets, QtCore
from traceback import print_exc
from paramiko.ssh_exception import AuthenticationException

from linien.client.config import load_device_data, save_device_data
from linien.client.widgets import CustomWidget
from linien.client.ui.new_device_dialog import Ui_NewDeviceDialog
from linien.client.connection import Connection
from linien.client.exceptions import GeneralConnectionErrorException, \
    ServerNotInstalledException, InvalidServerVersionException
from linien.client.dialogs import loading_dialog, error_dialog, execute_command, \
    question_dialog


class DeviceManagerCenter(QtGui.QWidget, CustomWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def ready(self):
        self.load_device_data(autoload=True)

    def load_device_data(self, autoload=False):
        devices = load_device_data()
        lst = self.ids.deviceList
        lst.clear()

        for device in devices:
            lst.addItem('%s (%s)' % (device['name'], device['host']))

        if autoload and len(devices) == 1:
            self.connect_to_device(devices[0])

    def connect(self):
        devices = load_device_data()

        if not devices:
            return

        self.connect_to_device(devices[self.get_list_index()])

    def connect_to_device(self, device):
        self.loading_dialog = loading_dialog(self, device['host'])

        def do():
            connected = False
            display_error = None
            display_question = None

            try:
                conn = Connection(device['host'], device['username'], device['password'])
                connected = True

            except ServerNotInstalledException:
                display_question = """The server is not yet installed on the device. Should it be installed?"""
                question_callback = lambda: \
                    self.install_linien_server(device)

            except InvalidServerVersionException as e:
                display_question = \
                    "The server version (%s) does not match the client (%s) version." \
                    "Should the corresponding server version be installed?" \
                    % (e.remote_version, e.client_version)
                question_callback = lambda v=e.client_version: \
                    self.install_linien_server(device, version=v)

            except AuthenticationException:
                display_error = 'Error at SSH authentication. ' \
                      'Check username and password and verify that you ' \
                      'don\'t have any offending SSH keys in your known ' \
                      'hosts file.'

            except GeneralConnectionErrorException:
                display_error = "Unable to connect to device."

            except Exception as e:
                print_exc()
                display_error = 'Exception occured when connecting to the device.'

            self.loading_dialog.hide()

            if display_error:
                error_dialog(self, display_error)

            elif display_question:
                if question_dialog(self, display_question):
                    question_callback()

            elif connected:
                self.app().connected(conn, conn.parameters, conn.control)

        QtCore.QTimer.singleShot(100, do)

    def install_linien_server(self, device, version=None):
        version_string = ''
        stop_server_command = ''

        if version is not None:
            version_string = '==' + version
            # stop server if another version of linien is installed
            stop_server_command = 'linien_stop_server;'

        self.ssh_command = execute_command(
            self, device['host'], device['username'], device['password'],
            (
                '%s; '
                'pip3 install linien-server%s; '
                'linien_install_requirements; '
            ) % (stop_server_command, version_string),
            lambda: self.connect_to_device(device)
        )

    def new_device(self):
        self.dialog = QtWidgets.QDialog()
        ui = Ui_NewDeviceDialog()
        ui.setupUi(self.dialog)
        self.dialog.setModal(True)
        self.dialog.show()

        def reload_device_data():
            # not very elegant...
            QtCore.QTimer.singleShot(100, self.load_device_data)

        self.dialog.accepted.connect(reload_device_data)

    def get_list_index(self):
        return self.ids.deviceList.currentIndex().row()

    def remove_device(self):
        devices = load_device_data()

        if not devices:
            return

        devices.pop(self.get_list_index())
        save_device_data(devices)
        self.load_device_data()

    def selected_device_changed(self):
        idx = self.get_list_index()

        disable_buttons = True

        if idx >= 0:
            devices = load_device_data()

            if devices:
                disable_buttons = False

        for btn in [self.ids.connectButton, self.ids.removeButton]:
            btn.setEnabled(not disable_buttons)


import paramiko

from PyQt5.QtWidgets import QApplication, QWidget, QListWidget, QVBoxLayout, QLabel, QPushButton, QListWidgetItem, \
    QHBoxLayout, QDialog, QMessageBox
from pyqtgraph import QtCore

from linien.client.utils import connect_ssh


class SSHCommandOutputWidget(QListWidget):
    def __init__(self, *args):
        super().__init__(*args)
        self.setSelectionMode(self.NoSelection)

    def execute(self, ssh, command, after_command):
        self.after_command = after_command
        self.stdin, self.stdout, self.stderr = ssh.exec_command(command)
        self.addItem('>>> %s' % command)
        self.show_output()

    def show_output(self):
        if self.stdout.channel.exit_status_ready():
            return self.after_command()
        else:
            for output in (self.stdout, self.stderr):
                buf = b''
                while output.channel.recv_ready():
                    buf += output.read(1)

                if buf:
                    self.addItem(buf.decode('utf8').rstrip('\n'))
                    self.scrollToBottom()

        QtCore.QTimer.singleShot(100, lambda: self.show_output())


def execute_command(parent, host, user, password, command, callback):
    ssh = connect_ssh(host, user, password)

    window = QDialog(parent)
    window.resize(800, 600)
    window_layout = QVBoxLayout(window)

    widget = SSHCommandOutputWidget(parent)
    window_layout.addWidget(widget)
    window.setLayout(window_layout)
    window.setModal(True)
    window.setWindowModality(QtCore.Qt.WindowModal)
    window.show()

    def after_command():
        window.hide()
        callback()

    widget.execute(ssh, command, callback)

    return window


def loading_dialog(parent, host):
    dialog = QMessageBox(parent)
    dialog.setIcon(QMessageBox.Information)
    dialog.setText('Connecting to %s' % host)
    dialog.setWindowTitle('Connecting')
    dialog.setModal(True)
    dialog.setWindowModality(QtCore.Qt.WindowModal)
    dialog.setStandardButtons(QMessageBox.NoButton)
    dialog.show()
    return dialog


def error_dialog(parent, error):
    return QMessageBox.question(
        parent, 'Error', error,
        QMessageBox.Ok,
        QMessageBox.Ok
    )


def question_dialog(parent, question):
    reply = QMessageBox.question(
        parent, 'Error', question,
        QMessageBox.Yes | QMessageBox.No,
        QMessageBox.Yes
    )

    return reply == QMessageBox.Yes


import os
import sys
import threading

from plumbum import colors

sys.path += ['../', '../../']

import linien
from linien.client.gui import QTApp


def run_application():
    print('Linien spectroscopy lock version ' + (colors.bold | linien.__version__))
    gui = QTApp()

    sys.exit(gui.app.exec_())


if __name__ == '__main__':
    run_application()
